!
! imke --- generate the potential of Mller et al.
!       from Int. J. Quantum Chem. *94* 75--92 (2003)
! Christian Buth
! 8 June 2006
!

      program imke
      use fortrantl
      implicit none

      integer i
      real(dblprec) r, rstep

      rstep = 30.0D0 / (441.0D0 - 1.0D0)

      do i=1, 441
        r = (i-1) * rstep
        ! Potential for plotting
!         write (*,*) r, Vimke(r), r * Vimke(r)
        ! rV for oneel; note that nl=2 calculations (i.e. l=1) are required to get the proper
        ! centrifugal potential
        write (*,100) r, redVimke(r)
      end do

 100  format(e25.16,e25.16)


      contains

!
! Vimke --- Imke Britta Mller's potential from Int. J. Quantum
!       Chem. *94* 75--92 (2003)
!

      real(dblprec) function Vimke(r)
      use fortrantl
      implicit none

      real(dblprec) r

      Vimke = 1.0D0 / r**2 - 4.0D0 / r + 7.0D0 * exp(-6.0D0 * (r - 6.0D0)**2)
      end function Vimke


!
! redVimke --- Imke Britta Mller's potential times r where
!       the l=1 centrifugal potential term was eliminated
!       beforehand
!

      real(dblprec) function redVimke(r)
      use fortrantl
      implicit none

      real(dblprec) r

      redVimke = -4.0D0 + 7.0D0 * r * exp(-6.0D0 * (r - 6.0D0)**2)
      end function redVimke
      end program imke
