!
! hamiltonian --- build the complete Hamiltonian matrix
! Robin Santra, Christian Buth
! 25 Sep 2006
!


      subroutine hamiltonian(dimen, nmain, nphmin, nphmax, lmin, lmax, nvec, mi,  &
     &          cconst, eval, Eground, omega, SESfTfatm, SESpotatm,        &
     &          SESWatom, SESdipatm, ceval)
      use floquetmat, hamiltoniandbl => hamiltonian
      use matrixfile
      implicit none

      integer, intent(in) :: dimen, nmain, nphmin, nphmax, lmin, lmax, nvec, mi
      real(dblprec), intent(in) :: cconst    ! Coupling constant sqrt(2 pi alpha intensity)
      real(dblprec), intent(in) :: eval(nvec,lmin:lmax)
      real(dblprec), intent(in) :: Eground, omega
      complex(dblprec), intent(in) :: SESfTfatm(nvec,nvec,lmin:lmax),  &
     &                                SESpotatm(nvec,nvec,lmin:lmax),  &
     &                                SESWatom (nvec,nvec,lmin:lmax),  &
     &                                SESdipatm(nvec,lmin:lmax,nvec,lmin:lmax)
      complex(dblprec), intent(in), optional :: ceval(nvec,lmin:lmax)

      integer indr, &                   ! Hamiltonian matrix row index
     &        indc, &                   ! Hamiltonian matrix column index
     &        mur,  &                   ! mu row index
     &        muc,  &                   ! mu column index
     &        lr,lc, i, j
      real(dblprec) scal
      complex(dblprec) z
      type(tmatrixfile) matf
      
      ! Open files for the diagonal and the off-diagonal matrix
      ! elements of the CO-ADC matrix
      matf%readfile     = .false.
      matf%fmtxcom      = 393
      matf%fmtxdia      = 394
      matf%fmtxoff      = 395
      matf%main         = nmain
      matf%ndim         = dimen
      matf%nbuf         = 10000
      matf%threshold    = 1.0D-14
      call openmatrixfile(matf)

      indc = 0                          ! Column index of Hamiltonian
      do muc = nphmin, nphmax
        do lc = lmin, lmax
          do j = 1, nvec
            indc = indc + 1
            indr = 0                    ! Row index of Hamiltonian
            do mur = nphmin, nphmax
              do lr = lmin, lmax
                do i = 1, nvec
                  indr = indr + 1

                  ! Only calculate the upper triangle of the Hamiltonian matrix
                  if (indr > indc) then
                    cycle
                  end if

                  ! Do coupling blocks; this is mur = muc-1
                  if (muc .gt. nphmin .and. mur == muc - 1) then
                    call calcscal(lr,lc,mi,scal)
                    scal = scal * cconst
                    z = cmplx(scal, kind=dblprec) * SESdipatm(i,lr,j,lc)
                    call writeoffdiagentry(matf, indr, indc, z)
                  end if

                  if (muc == mur .and. lc == lr) then
                    z =     SESfTfatm(i,j,lc)
                    z = z + SESpotatm(i,j,lc)
                    z = z + SESWatom (i,j,lc)
                    ! The full diagonal mur = muc, lr = lc and i = j
                    ! Here indr == indc holds
                    if (i == j) then
                      z = z + cmplx(-Eground - real(muc, dblprec) * omega, kind=dblprec)
                      if (present(ceval)) then
                        z = z + ceval(j,lc)
                      else
                        z = z + cmplx(eval(j,lc), kind=dblprec)
                      end if
                      call writediagentry(matf, indr, z)

                    else
                      call writeoffdiagentry(matf, indr, indc, z)
                    end if
                  end if

                end do
              end do
            end do
          end do
        end do
      end do

      call closematrixfile(matf)


      contains

!
! calcscal --- calculate the Clebsch-Gordon coefficients for the
!       dipole operator
!
! lf : final angular momentum
! li : initial angular momentum
! mi : initial magnetic quantum number
! scal : return value
!

      subroutine calcscal(lf,li,mi,scal)
      use anglib
      implicit none

      integer lf,li,mi
      real(dblprec) scal

      scal = dsqrt(dble(2*li+1)/dble(2*lf+1))
      scal = scal*cleb(2*li,2*mi,2,0,2*lf,2*mi)
      scal = scal*cleb(2*li,0,2,0,2*lf,0)
      end subroutine calcscal
      end subroutine hamiltonian
