!
! Vmtrx --- represent the potential in terms of finite element
!       basis functions
! Kevin Christ, Mark Baertschy, Christian Buth
! 9 June 2006
!
! Version 1.0.0: 21 Aug 2006 (The potential points are now generated
!       in a separate subroutine)
!
! ll: angular momentum quantum number for the potential l (l + 1) / (2 * r)
!


      subroutine Vmtrx(V,ll,rr,rV,b,c,d,x,ipottype)
      use fortrantl
      use FiniteElements
      implicit none

!       type(FEBnodes) :: x
!       real(dblprec) :: V(-5:5,*)
! CB
      type(FEBnodes), intent(in) :: x
      real(dblprec), intent(inout) :: V(-5:5,1:numbasfun(x))
! END CB

  real(dblprec) ll, rr(npotentialpoints), rV(npotentialpoints), b(npotentialpoints), c(npotentialpoints), d(npotentialpoints)
  integer nmnds, nbasis, left, right, gnm, row1, j, k, l, mu, ipottype
  REAL(dblprec), ALLOCATABLE :: f(:), r(:), gx(:), gw(:), q(:,:), p(:,:)
  REAL(dblprec) :: h1, h2
!   REAL(dblprec) :: tol, seval

  nmnds = x%numnds
  left = x%numleft
  right = x%numright

!!numbasfun is defined in FiniteElements.f90 and gives the total # of basis functions
!!nbasis will be the size (nbasis*nbasis) of the full V matrix

  nbasis = numbasfun(x)
  IF (nbasis <= 0) THEN
    RETURN
  ENDIF
  V(-5:5,1:nbasis) = 0.d0

!   tol = 1.D-5

  gnm = 20  !!This is the amount of points to be used by the Gaussian integration

  ALLOCATE ( r(gnm), f(gnm), gx(gnm), gw(gnm), q(gnm,0:2), p(gnm,0:2) )

  CALL gauleg(-1.d0, 1.d0, gx, gw, gnm)

!!Coordinate transformation for Gaussian integration on interval (0,1)


  gx(:) = (gx(:) + 1.d0)/2.d0
  gw(:) = gw(:)/2.d0

  r(:) = gx(:)

  DO mu = 0,2
    q(:,mu) = FEbasfun(r,mu)
  END DO
  r(:) = r(gnm:1:-1)
! Buth
! gfortran has problems here
!   DO mu = 0,2
!     p(:,mu) = ((-1)**mu)*FEbasfun(r,mu)
!   END DO
      do mu = 0,2
        do j=1, gnm
          p(j,mu) = ((-1)**mu)*FEbasfun(r(j),mu)
        end do
      end do
! END Buth

      ! half-basis functions "centered" at nodes(1)
      row1 = left-2
      h2 = node_spacing(x,1)
      r(:) = h2*gx(:) + x%nodes(1)
!       ! First: At this place you can insert the anlytic form of
!       ! a potential instead of reading a point mesh and
!       ! using spline interpolation
!       do j = 1, gnm
!         ! Potential of Bain et al. from J. Phys. B *26* 4503 (1993)
!         f(j) = 7.5D0 * r(j)**2 * exp(-r(j))
!         ! Spline interpolated potential
! !      f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 + seval(npotentialpoints,r(j),rr,rV,b,c,d)/r(j)
!       end do
      call potpoints(gnm, r, f, ipottype)

  DO j = (3-left),2
    DO k = j,2
      V(j-k,row1+k) = (h2**(j+k+1))*SUM(gw(:)*p(:,j)*p(:,k)*f(:))
    END DO
  END DO


! overlaps between nodes(1) and nodes(2)
  h1 = h2
  IF (nmnds > 2) THEN
    DO j = (3-left),2
      DO k = 0,2
        V(j-k-3,row1+3+k) = (h1**(j+k+1))*sum(gw(:)*p(:,j)*q(:,k)*f(:))
      END DO
    END DO
  ELSE
    DO j = (3-left),2
      DO k = (3-right),2
        V(j-k-left,row1+left+k) = (h1**(j+k+1))*sum(gw(:)*p(:,j)*q(:,k)*f(:));
      END DO
    END DO
  ENDIF

  h1 = h2;
! basis functions centered at nodes(2)
  IF (nmnds > 2) THEN
    row1 = left + 1
    DO j = 0,2
      DO k = j,2
        V(j-k,row1+k) = (h1**(j+k+1))*sum(gw(:)*q(:,j)*q(:,k)*f(:))
      END DO
    END DO
    h2 = node_spacing(x,2)
    r(:) = h2*gx(:) + x%nodes(2)

!     DO j = 1, gnm
! ! Second: At this place you can insert the anlytic form of
! ! a potential instead of reading a point mesh and
! ! using spline interpolation
! !       f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 - 1.d0/(cosh(r(j)))**2 - tanh(r(j))/r(j)
! !       f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 - 1.d0/r(j)
!        f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 + seval(npotentialpoints,r(j),rr,rV,b,c,d)/r(j)
!     END DO
      call potpoints(gnm, r, f, ipottype)

    DO j = 0,2
      DO k = j,2
        V(j-k,row1+k) = V(j-k,row1+k) + (h2**(j+k+1))*sum(gw(:)*p(:,j)*p(:,k)*f(:))
      END DO
    END DO
  ENDIF

  DO l = 3,(nmnds-1)
! overlaps between nodes(l-1) and nodes(l)
    h1 = h2
    DO j = 0,2
      DO k = 0,2
        V(j-k-3,row1+3+k) = (h1**(j+k+1))*sum(gw(:)*p(:,j)*q(:,k)*f(:))
      END DO
    END DO

  ! basis functions centered on nodes(l)
    row1 = row1 + 3;
    DO j = 0,2
      DO k = j,2
        V(j-k,row1+k) = (h1**(j+k+1))*sum(gw(:)*q(:,j)*q(:,k)*f(:))
      END DO
    END DO
    h2 = node_spacing(x,l)
    r(:) = h2*gx(:) + x%nodes(l)

!     DO j = 1, gnm
! ! Third: At this place you can insert the anlytic form of
! ! a potential instead of reading a point mesh and
! ! using spline interpolation
! !       f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 - 1.d0/(cosh(r(j)))**2 - tanh(r(j))/r(j)
! !       f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 - 1.d0/r(j)
!        f(j) = 0.5d0*ll*(ll+1.d0)/r(j)**2 + seval(npotentialpoints,r(j),rr,rV,b,c,d)/r(j)
!     END DO
      call potpoints(gnm, r, f, ipottype)


    DO j = 0,2
      DO k = j,2
        V(j-k,row1+k) = V(j-k,row1+k) + (h2**(j+k+1))*sum(gw(:)*p(:,j)*p(:,k)*f(:))
      END DO
    END DO
  END DO

! overlaps between nodes(nmnds-1) and nodes(nmnds)
  IF (nmnds > 2) THEN
    h1 = h2
    DO j = 0,2
      DO k = (3-right),2
        V(j-k-right,row1+right+k) = (h1**(j+k+1))*sum(gw(:)*p(:,j)*q(:,k)*f(:))
      END DO
    END DO
  ENDIF

! half-basis functions "centered" on nodes(nmnds)
  row1 = row1 + right
  DO j = (3-right),2
    DO k = j,2
      V(j-k,row1+k) = (h1**(j+k+1))*sum(gw(:)*q(:,j)*q(:,k)*f(:))
    END DO
  END DO

  DO k = 1,(nbasis-1)
    DO j = (k+1),MIN(nbasis,k+5)
      V(j-k,k) = V(k-j,j)
    END DO
  END DO


      contains

!
! potpoints --- determine the potential for an array of points
!

      subroutine potpoints(gnm, r, f, ipottype)
      use fortrantl
      implicit none

      integer, intent(in) :: gnm, ipottype
      real(dblprec), intent(in) :: r(:)
      real(dblprec), intent(out) :: f(:)

      logical, save :: message = .false.
      integer j
      real(dblprec) seval
      external seval

      if (.not. message) then
        select case (ipottype)
          case (0)
            write (*,*) 'Using the spline interpolated potential given on a grid'
          case (1)
            write (*,*) 'Using the potential of Bain et al. from J. Phys. B *26* 4503 (1993)'
          case (2)
            write (*,*) 'Using the potential of Mller et al. from Int. J. Quantum Chem. *94* 75--92 (2003)'
          case default
            stop 'Vmtrx: unknown potential requested!'
        end select
        message = .true.
      end if

      ! At this place you can insert the anlytic form of
      ! a potential instead of using a point mesh in conjunction
      ! with spline interpolation
      do j = 1, gnm
        select case (ipottype)
          case (0)
            ! Spline interpolated potential
            f(j) = 0.5D0 * ll * (ll+1.0D0) / r(j)**2  &
     &          + seval(npotentialpoints, r(j), rr, rV, b, c, d) / r(j)
          case (1)
            ! Potential of Bain et al. from J. Phys. B *26* 4503 (1993)
            f(j) = 7.5D0 * r(j)**2 * exp(-r(j))
          case (2)
            ! Imke Britta Mller's potential from Int. J. Quantum
            ! Chem. *94* 75--92 (2003) where the l=1
            ! centrifugal potential term was replaced by the general
            ! term with l = ll.  Imke's results are obtained for
            ! calculations which include l=1.
            f(j) = 0.5D0 * ll * (ll+1.0D0) / r(j)**2  &
     &             -4.0D0/r(j) + 7.0D0*exp(-6.0D0 * (r(j) - 6.0D0)**2)
          case default
            stop 'Vmtrx: unknown potential requested!'
        end select
      end do
      end subroutine potpoints
      end subroutine Vmtrx
