! Fifth Order Finite Element function evaluation
!
! f is the function, evaluated at the
! points in the vector r, defined by expansion coefficients coef
! in terms of fifth-order polynomial finite elements for
! subintervals defined by nodes.
!
! this is for a radial grid (0 to Rmax)
! half basis functions of types 1 and 2 exist on the first and last intervals
!
!
! requirements:
!    coef = vector of expansion coefficients
!    nodes = data-structure defining finite element nodes
!    r = vector of grid points to evaluate the function


SUBROUTINE fxneval1(f,r,ngrd,coef,nodes)
  USE FiniteElements
  implicit none

  INTEGER, intent(IN) :: ngrd
  TYPE(FEBnodes), intent(IN) :: nodes
  REAL(kind=8), intent(OUT) :: f(ngrd)
  REAL(kind=8), intent(IN) :: r(ngrd), coef(nodes%numbasis)
  INTEGER :: left, right, nbasis, nodeptr, coptr, j, k
  REAL(kind=8) :: h, xp, xq

  f(:) = 0.d0

! number of half-basis functions for left and right ends
! default value is two for both.
  left = nodes%numleft
  right = nodes%numright

  nbasis = nodes%numbasis
  if ( nbasis==0 ) RETURN

  nodeptr = 1
  coptr = 0
  DO j = 1,ngrd
    if (r(j) >= nodes%nodes(nodeptr)) EXIT
  END DO

  coptr = 1
  DO j = j,ngrd
    IF (r(j)==nodes%nodes(nodeptr)) THEN
      IF (left < 3) THEN
        f(j) = 0
      ELSE
        f(j) = coef(coptr)
      ENDIF
    ELSE
      EXIT
    ENDIF
  END DO

  nodeptr = nodeptr + 1
  h = nodes%nodes(nodeptr) - nodes%nodes(nodeptr-1)
  coptr = left-2

  IF (nodes%numnds > 2) THEN
    k = ngrd + 1
    DO j=j,ngrd
      IF (r(j) >= nodes%nodes(nodeptr)) THEN
        k = j - 1
        EXIT
      ENDIF
      xp = (nodes%nodes(nodeptr)-r(j))/h
      xq = (r(j)-nodes%nodes(nodeptr-1))/h
      DO k = (3-left),2
        f(j) = f(j) + coef(coptr+k)*((-h)**k)*FEbasfun(xp,k)
      END DO
      DO k = 0,2
        f(j) = f(j) + coef(coptr+3+k)*((h)**k)*FEbasfun(xq,k)
      END DO
!! HACK!! FIX THIS!!
      k = j
    END DO
    j = k
  ELSE
    DO j = j,ngrd
      if (r(j) >= nodes%nodes(nodeptr)) EXIT
      xp = (nodes%nodes(nodeptr)-r(j))/h
      xq = (r(j)-nodes%nodes(nodeptr-1))/h
      DO k = (3-left),2
        f(j) = f(j) + coef(coptr+k)*((-h)**k)*FEbasfun(xp,k)
      END DO
      DO k = (3-right),2
        f(j) = f(j) + coef(coptr+right+k)*((h)**k)*FEbasfun(xq,k)
      END DO
    END DO
  ENDIF

  j = j + 1
  DO j = j,ngrd
    IF (r(j) > nodes%nodes(nodeptr)) THEN
      nodeptr = nodeptr + 1
      coptr = coptr + 3
      IF (nodeptr >= nodes%numnds) THEN
!! hack!! fix this!!
        k = j - 1
        EXIT
      ENDIF
      h = nodes%nodes(nodeptr) - nodes%nodes(nodeptr-1)
    ENDIF
    IF (r(j) == nodes%nodes(nodeptr)) THEN
      f(j) = coef(coptr+3)
    ELSE
      xp = (nodes%nodes(nodeptr)-r(j))/h
      xq = (r(j)-nodes%nodes(nodeptr-1))/h
      DO k = 0,2
        f(j) = f(j) + coef(coptr+k)*((-h)**k)*FEbasfun(xp,k)
      END DO
      DO k = 0,2
        f(j) = f(j) + coef(coptr+3+k)*((h)**k)*FEbasfun(xq,k)
      END DO
    ENDIF
!! hack!! fix this!!
    k = j
  END DO

  if (nodeptr > nodes%numnds) RETURN

  j = k + 1
  DO j = j,ngrd
    if (r(j) > nodes%nodes(nodeptr)) RETURN
    IF (r(j) == nodes%nodes(nodeptr)) THEN
      if (right == 3) f(j) = coef(coptr+3)
    ELSE
      xp = (nodes%nodes(nodeptr)-r(j))/h
      xq = (r(j)-nodes%nodes(nodeptr-1))/h
      DO k = 0,2
        f(j) = f(j) + coef(coptr+k)*((-h)**k)*FEbasfun(xp,k)
      END DO
      DO k = (3-right),2
        f(j) = f(j) + coef(coptr+right+k)*((h)**k)*FEbasfun(xq,k)
      END DO
    ENDIF
  END DO

  RETURN
END SUBROUTINE fxneval1


SUBROUTINE fxneval2(f,r,ngrd,coef,nodes)
  USE FiniteElements
  implicit none

  INTEGER :: ngrd
  TYPE(FEBnodes) :: nodes
  REAL(kind=8) :: f(ngrd,ngrd), r(ngrd), coef(nodes%numbasis,nodes%numbasis)
  INTEGER :: nbasis, j, k
  REAL(kind=8) :: h, xp, xq
  REAL(kind=8), ALLOCATABLE :: temp(:,:), coef1(:)

  f(:,:) = 0.d0
  nbasis = nodes%numbasis
  ALLOCATE( temp(ngrd,nbasis), coef1(nbasis) )
  DO j = 1,nbasis
    CALL fxneval1(temp(1,j),r,ngrd,coef(1,j),nodes)
  END DO
  DO j = 1,ngrd
    coef1(:) = temp(j,:)
    CALL fxneval1(f(1,j),r,ngrd,coef1,nodes)
  END DO

  RETURN
END SUBROUTINE fxneval2
